#include <avr/io.h>
#include <stdint.h>
#include <stdlib.h>
#include "i2c.h"
#include "usart.h"
#include "mmc_lib.h"
#include "spi.h"
#include "other.h"
#include "registerstrings.h"
#include "ssc.h"
#include "sdc.h"
#include "fat16.h"
#include "hd44780.h"
#include <avr/pgmspace.h>
#include <string.h>
#include "joystick.h"
#define TIMEOUT_WAIT		0xFFFFFFUL	// So hoch wird gezhlt, bis Timeout erreicht wird

char registernames[40][17];				// Definiere Strings (Dekleration in registerstrings.h
uint8_t used_register_names[40];		// Beides Global damit berall Einsetzbar (extern definiert)
uint8_t errorcode;						// Definition von Errorcode (Global definiert)
uint32_t timeout_counter;				// Definition von Timeoutzhlervariable (Global definiert)
uint8_t resetbit;						// Gibt an, ob zurckgesetzt werden soll(1=> Rcksetzen)

/*******************************************************
	Funktion :		Main
	Eigenschaften:	Hauptprogramm
	bergabewert:	-
	Rckgabewert:	-
********************************************************/
int main(void) 
{
	struct {											// Bitfeld fr Flags
		unsigned live_active:1;
		unsigned send_all_active:1;
		unsigned receive_ack:1;
		unsigned other_packet_receive:1;				// Da es 2 packet_receive Abfragen gibt, die 2te aber nur fr ACK ist
														// darf wenn bei der 2ten Abfrage ein anderes Packet als ACK empfangen
														// wird die erste Abfrage nicht noch ein Packet holen, sondern das von
														// der 2ten benutzen. Dies ist der Fall wenn other_packet_receive = 1 ist
		unsigned timeout:1;								// timeout = 1 (timeout erreicht), timeout = 0 (kein timeout)
	} flag;

	struct time pc, rtc, start, end;
		
	uint8_t 	temp_seconds_last	= 0x00;				// Zwischenspeicher fr Sekunden (letzte Sekunde abspeichern)
	uint8_t 	temp_seconds_real	= 0x00;				// Zwischenspeicher fr Sekunden (aktuelle Sekunde abspeichern)
	uint32_t 	sampling_rate;							// Sampling Rate
	uint32_t 	measurement_mask_low;					// Messwertemaske der niedrige Teil
	uint32_t 	measurement_mask_high;					// Messwertemaske der hohe Teil
	uint32_t 	measured[40];							// Array zum Speichern von Messdaten
	uint8_t 	measured_length;						// Anzahl der zu messenden Daten
	uint8_t 	command_nr 			= COMMAND_ZERO;		// Gibt ab welcher Befehl gerade aktiv ist
	uint8_t 	command_nr_live		= COMMAND_ZERO;		// Variable, die im Live Modus benutzt wird.
	uint8_t 	command_nr_live2	= COMMAND_ZERO;		// Variable, die im Live Modus benutzt wird	
	uint8_t		packet_counter = 0x00;					// Wird im Live Modus benutzt, um zu zhlen, wieoft Packet nicht 

	reset:												// Hierher wird gesprungen, wenn C rckgesetzt wird

	// Zeit von PC, in BCD eingabe
	pc.seconds 			= 0;				
	pc.minutes 			= 58;				
	pc.hours 			= 16;				
	pc.date 			= 7;				
	pc.month 			= 1;			
	pc.year 			= 9;
	
	// Startzeit der Messung, in BCD
	start.seconds		= 0;
	start.minutes		= 0;				
	start.hours			= 0;				
	start.date			= 0;				
	start.month			= 0;				
	start.year			= 0;				

	// Endzeit der Messung, in BCD
	end.seconds			= 0;				
	end.minutes			= 59;				
	end.hours			= 16;				
	end.date			= 7;				
	end.month			= 1;				
	end.year			= 9;				

	errorcode 			= EC_NO_ERROR;	// Errorcode Variable initialisieren
	timeout_counter			= 0;		// Timeoutzhlervariable initialisieren
	flag.live_active 		= 0;		// 0 = Live deaktiviert, 1 = Live aktiviert
	flag.send_all_active 	= 0;		// 0 = Schicke alle Daten deaktiviert, 1 = aktiviert
	flag.receive_ack 		= 0;		// 0 = kein ACK, 1 = ACK
	measurement_mask_low 	= 0b110UL;	// Anfangsmaske einstellen
	measurement_mask_high 	= 0b11UL;	// Anfangsmaske einstellen

// ---------------------------------------------------------------------------------------------------
// -------------------------------- Warte bis alle Gerte bereit sind --------------------------------
// ---------------------------------------------------------------------------------------------------
/*
		while(timeout_counter != TIMEOUT_WAIT)
		{
			timeout_counter++;
		}
		timeout_counter = 0;
*/

// ---------------------------------------------------------------------------------------------------
// ----------------------------------------- Initialisierung -----------------------------------------
// ---------------------------------------------------------------------------------------------------
	lcd_init();							// Initialisiere Display (an erster Stelle fr Fehlercodesanzeige)

	usart_init();						// Initialisiere USART fr USB-Kommunikation
	i2c_init();							// Initialisiere I2C fr RTC
	
	spi_init();							// Initialisiere SPI fr SDCard+Messchip
	registerstrings_init();				// Schreibe Registernamen in Array registernames

	i2c_send_time(pc);					// Stelle RTC-Register auf Anfangszeit

	spi_select_measuring_chip();		// Aktiviere Messchip
	spi_send_start_conversions();		// Starte Messungen
	spi_unselect();						// Deaktivere Messchip
	
	joystick_init();					// Initialisiere den Joystick


// ---------------------------------------------------------------------------------------------------
// ------------------------------------ Abfrage von RTC und USART ------------------------------------
// ---------------------------------------------------------------------------------------------------

	while(1)							// Endlosschleife fr zyklisches Arbeiten
	{
		while(temp_seconds_last == temp_seconds_real)	// Abfrage, ist eine Sekunde vergangen?
		{
			rtc = i2c_get_time();						// Hole Zeit um zu vergleichen
			temp_seconds_real = rtc.seconds;			// Speichern der Seconds (wegen BCD Umwandlung)

			// Hole die Command Nr.
			if(command_nr_live == COMMAND_ZERO && command_nr_live2 == COMMAND_ZERO)	// Kein berhang aus Live-Modus
			{
				if(usart_check())						// Neue Daten im USART-Speicher?
				{	
					command_nr = usart_recieve_packet();// Speichere den Befehlscode ab
				}
			}
			else if(command_nr_live2 == COMMAND_ZERO)	// 1 Packet berhang aus Live-Modus
			{
				command_nr = command_nr_live;			// Speichere den Befehlscode ab
				command_nr_live = COMMAND_ZERO;			// Zurcksetzen fr nchsten Schleifedurchlauf
			}
			else										// 2 Packete berhang aus Live-Modus
			{
				command_nr = command_nr_live2;			// Speichere den Befehlscode ab
				command_nr_live2 = COMMAND_ZERO;		// Zurcksetzen, fr nchsten Schleifendurchlauf
			}

			switch(command_nr)							// Welchen Befehl ausfhren?
			{
				case COMMAND_ZERO:
				{
					// Tue nichts, damit bei command_nr == 0x00 nicht default geht
				} break;
				case COMMAND_SEND_ALL_DATA:				// Schicke alle Daten
				{
					flag.send_all_active = 1;			// Aktiviere Modus "Schicke alle Daten"
					command_nr = COMMAND_ZERO;			// Command_nr zurcksetzen, da Befehl fertig
				} break;

				case COMMAND_LIVE_MODUS:				// Live Modus
				{
					flag.live_active = 1;				// Live Modus aktivieren
					command_nr = COMMAND_ZERO;			// Command_nr zurcksetzen, da Befehl fertig
				} break;

				case COMMAND_LIVE_MODUS_STOP:			// Live Modus Stopp
				{
					flag.live_active = 0;				// Live Modus deaktivieren
					command_nr = COMMAND_ZERO;			// Command_nr zurcksetzen, da Befehl fertig
				} break;
	
				case COMMAND_RECEIVE_TIME:				// Uhrzeit empfangen
				{
					if(usart_receive_time(&pc)) 		// Daten richtig? + Hole die Geschickte Uhrzeit
					{
						usart_send_ack();				// Sende ACK, da Daten richtig empfangen
						i2c_send_time(pc); 				// Sende Uhrzeit an RTC
						command_nr = COMMAND_ZERO;		// Setze Command_nr zurck, da Befehl fertig
					}
				} break;

				case COMMAND_RECEIVE_ACK:				// Receive ACK
				{
						flag.receive_ack = 1;			// Setze Flag
						command_nr = COMMAND_ZERO;		// Setze Command_nr zurck, da Befehl fertig
				} break;

				case COMMAND_PARAMETERS:				// Parameters
				{
					if(usart_parameters(&start, &end, &sampling_rate, &measurement_mask_low, &measurement_mask_high )) 
					{									// Hole die geschickten Paramter + berprfung ob richtig
						usart_send_ack();				// Sende ein ACK
						command_nr = COMMAND_ZERO;		// Setze Command_nr zurck, da Befehl fertig
					}
				} break;
				
				case COMMAND_STATUS_REQUEST:			// Status-senden Aufforderung
				{
					usart_status_response(errorcode);
					command_nr = COMMAND_ZERO;			// Setze Command_nr zurck, da Befehl fertig
				} break;

				case COMMAND_RESET:
				{
					resetbit = 1;						// Setze das Flag
					command_nr = COMMAND_ZERO;
				} break;
			}									// Ende Switch


			if(flag.send_all_active == 1)		// Schicke alle Daten Modus aktiv? (steht in While da so schneller)
			{

												// Hole 1te Datei aus datein.txt
												// Ist zur Zeit eine Datei geffnet?
												// Wenn ja, hole Daten von nchster Zeile

												// Schaue ob ACK erhalten?
												// ACK zurcksetzen nicht vergessen!

												// Schicke Daten ber USART
												// erhhe nchste Zeile
												// keine Neue Datei offen.... ffne

			}

			if(resetbit == 1)
			{
				resetbit = 0;
				usart_seq_reset();
				goto reset;
			}
			// Joystick Taste Abfragen
			joystick_measure(measured_length,errorcode, rtc, measurement_mask_low, measurement_mask_high, measured);

		}										// Ende While
		temp_seconds_last = rtc.seconds;		// Speichere fr nchste Schleifenberprfung

		
// ---------------------------------------------------------------------------------------------------
// ----------------------------------- Hole Messwerte von Messchip -----------------------------------
// ---------------------------------------------------------------------------------------------------

		spi_select_measuring_chip();
		spi_get_measured(&measured_length, &measurement_mask_low, &measurement_mask_high, measured);
		spi_unselect();


// ---------------------------------------------------------------------------------------------------
// --------------------------------------- Speichere die Werte ---------------------------------------
// ---------------------------------------------------------------------------------------------------
/*
		if(time_in_target_area(rtc,start,end))
		{
			for(uint8_t elementcounter = 0; elementcounter < measured_length; ++elementcounter) // Durchlaufe alle gemessenen Werte
			{
				//usart_send_24bit(measured[elementcounter]);
				//usart_send(0xFF);
			}
		}
*/

// ---------------------------------------------------------------------------------------------------
// ------------------------------------- Modus: Live bertragung -------------------------------------
// ---------------------------------------------------------------------------------------------------

		if(flag.live_active == 1)								// Live bertragung aktiv?
		{
			flag.other_packet_receive = 0;						// Sicherheitshalber auf 0 zurcksetzen
			packet_counter = 0;									// Sicherheitshalber auf 0 zurcksetzen

			live_send:											// Goto Sprungmarke (aufpassen!)
			usart_new_line(measured_length,measured,rtc);		// Schicke "Neue Zeile"

			live_check:											// Goto Sprungmarke (aufpassen!)
			timeout_counter = 0;								// damit nach Sprung nichtmehr gesetzt ist
			flag.timeout = 0;									// damit nach Sprung nichtmehr gesetzt ist

			while(timeout_counter <= TIMEOUT_WAIT)
			{
				if(usart_check())						// Neue Daten im USART-Speicher?
				{	
					flag.timeout = 0;
					break;	
				}
				else
				{
					timeout_counter++;
					flag.timeout = 1;
				}
			}

			if(flag.timeout == 1)								// Gab es ein Timeout?
			{
				if(flag.other_packet_receive == 1)				// 1. Packet falsch 2. Timeout
				{
					flag.live_active = 0;						// Live Modus wegen Fehler beenden
					flag.other_packet_receive = 0;				// Zurcksetzen, da ERROR
					programm_error(EC_LIVE_WRONG_PACKET_TO);	// Error-Routine aufrufen
				}
				else											// Es gab bis jetzt nur Timeouts
				{
					if(packet_counter >= 2)						// Wurden Daten schon 3 mal falsch gesendet?
					{	
						flag.live_active = 0;					// Live Modus wegen Fehler beenden
						flag.other_packet_receive = 0;			// Zurcksetzen, da ERROR
						programm_error(EC_LIVE_TIMEOUT);		// Error-Routine aufrufen
					}
					else										// Weniger als 3 falsche packet erhalte
					{
						packet_counter++;						// Erhhe Fehleranzahl um 1
						usart_seq_minus();						// Weil gleiches Packet nocheinmal gesendet wird
						goto live_send;							// Schicke Daten nocheinmal
					}
				}
			}
			else												// Es wurde ein Packet erhalten (Kein Timeout)
			{	
				if(flag.other_packet_receive == 0)				// Erstes Packet
				{
					command_nr_live = usart_recieve_packet();	// Hole den Befehlscode
					command_nr = command_nr_live;
				}
				else											// Es wurde schon ein falsches Packet erhalten
				{
					command_nr_live2 = usart_recieve_packet();	// Hole den Befehlscode
					command_nr = command_nr_live2;
				}
				
				if(COMMAND_RECEIVE_ACK == command_nr)			// Wurde richtiges Packet erhalten? (ACK)
				{
					if(flag.other_packet_receive == 0)			
					{
						command_nr_live = 0;					// War ein ACK und wurde erledigt
					}
					else									
					{	
						command_nr_live2 = 0;					// War ein ACK und wurde erledigt
					}
					flag.other_packet_receive = 0;				// Richtiges Packet erhalten
					command_nr = COMMAND_ZERO;					// Setze die Command Nr zurck
				}
				else if(COMMAND_LIVE_MODUS_STOP == command_nr)	// Wurde ein Stopp Packet gesendet?
				{
					// Nicht mehr auf ACK warten
					// Tue nichts und verlasse einfach Live-Modus
					// In oberen While-Schleife wird Live-Modus beendet
				}
				else
				{
					if (flag.other_packet_receive == 1)			// Ist das schon das 2te falsche Packet?
					{
						flag.live_active = 0;					// Live Modus wegen Fehler beenden
						flag.other_packet_receive = 0;			// Zurcksetzen, da ERROR
						programm_error(EC_LIVE_WRONG_PACKAGES); // Error-Routine aufrufen
					}
					else
					{
						flag.other_packet_receive = 1;			// Falsches Packet erhalten
						goto live_check;						// Hole noch ein Packet
					}
				}		// Ende If() ob richtiges Packet(ACK)
			}			// Ende If() ob Timeout war
		}				// Ende If() Live Modus
	}					// Ende Endlosschleife
}						// Ende Main-Funktion
